
# Note: List of supported cycles material node types in node_wrangler.py

import bpy
import os, sys
import shutil

# debugging
# import pdb
import winsound

from mathutils import Vector
from .util import Util
from .file import File


# Diagnostics output colors :

Warning = "34;40" # Blue/Black
Info = "34;40" # Blue/Black
Error ="31;40" # Red/Black


class Export_Material:
	
	@staticmethod
	def Write(exp, mtl):

		exp.file.Indent()

		if not mtl:  
			# Default material for anything undefined. 
			exp.file.Write('<material id="default"><part id="diffuse"><input color="0.5 0.5 0.5"/></part></material>\n')
			exp.file.Unindent()
			return
			
		exp.file.Write('<material id="{}"'.format(Util.SafeName(mtl.name)))
        
		normal_map_input = None
		         
		# Indicates we've written <material [attribs]> & are now ready to begin writing material parts. Required so we don't try to close an element tag more than once.
		mtl_open_complete = False
		
		if not mtl.node_tree:
			exp.file.Write('">\n', Indent=False)
			exp.file.Indent()
			
			# Diffuse
			clr = mtl.diffuse_color * mtl.diffuse_intensity
			exp.file.Write('<part id="diffuse"><input color="{:f} {:f} {:f}"/></>\n'.format(clr[0], clr[1], clr[2]))

			# Specular
			clr = mtl.specular_color * mtl.specular_intensity
			
			if (clr[0] > 0) or (clr[0] > 0) or (clr[2] > 0) :
				exp.file.Write('<part id="specular"><input color="{:f} {:f} {:f}"/></>\n'.format(clr[0], clr[1], clr[2]))
		else:
		
			# Principled BSDF conversion :

			principled_bsdf = mtl.node_tree.nodes.get("Principled BSDF")
			
			if principled_bsdf:
			
				alpha_input = principled_bsdf.inputs['Alpha']
				
				if not alpha_input.is_linked:
					alpha_value = alpha_input.default_value
					if alpha_value != 1.0:
						exp.file.Write(' opacity="{:f}"'.format(alpha_value))
            
				exp.file.Write('>\n', Indent=False)
				exp.file.Indent()
				mtl_open_complete = True

				# Not currently supported XSG like this, but gets it to file for consideration.
				if alpha_input.is_linked:
					Export_Material.Part_Write(exp, mtl, "opacity", alpha_input, -1) 

				diffuse = principled_bsdf.inputs[0]
				skip = False				
				
				if not diffuse.is_linked:			
					color = diffuse.default_value
					skip = (color[0] == 0) and (color[1] == 0) and (color[2] == 0)
	
				if not skip:
					Export_Material.Part_Write(exp, mtl, "diffuse", diffuse, -1)
					
				if principled_bsdf.inputs['Specular Tint'].is_linked:
					Export_Material.Part_Write(exp, mtl, "reflection", principled_bsdf.inputs['Specular Tint'], -1)
					
				# We use self illum as constant channel for now
				constant_input = principled_bsdf.inputs['Emission Color']
				skip = False				
				
				if not constant_input.is_linked:			
					color = constant_input.default_value
					emit_strength = principled_bsdf.inputs['Emission Strength'].default_value
					color[0] = color[0] * emit_strength
					color[1] = color[1] * emit_strength
					color[2] = color[2] * emit_strength
					skip = (color[0] == 0) and (color[1] == 0) and (color[2] == 0)
		
				if not skip:
					Export_Material.Part_Write(exp, mtl, "constant", constant_input, -1)
				
				normal_map_input = principled_bsdf.inputs['Normal']
				
			# Refraction BSDF 
			refraction_bsdf = mtl.node_tree.nodes.get("Refraction BSDF")
			
			if refraction_bsdf:
				Export_Material.Part_Write(exp, mtl, "refract", refraction_bsdf.inputs['Color'], refraction_bsdf.inputs['IOR'].default_value)


			# Diffuse BSDF : incorporates normal & diffuse
			diffuse_bsdf = mtl.node_tree.nodes.get("Diffuse BSDF")
			
			if diffuse_bsdf:

				if not mtl_open_complete:
					exp.file.Write('>\n', Indent=False)
					exp.file.Indent()
					mtl_open_complete = True
                
				Export_Material.Part_Write(exp, mtl, "diffuse", diffuse_bsdf.inputs[0], -1)
				normal_map_input = diffuse_bsdf.inputs[2]
			
			# Gloss / specular
			glossy_bsdf = mtl.node_tree.nodes.get("Glossy BSDF")
			
			if glossy_bsdf:
				Export_Material.Part_Write(exp, mtl, "specular", glossy_bsdf.inputs[0], glossy_bsdf.inputs[1].default_value * 1000)
			
			# Export Emission
			emission_part = mtl.node_tree.nodes.get("Emission")
			
			if emission_part:
				if mtl_open_complete != True:
					exp.file.Write('>\n', Indent=False)
					exp.file.Indent()
					mtl_open_complete = True
                                    
				Export_Material.Part_Write(exp, mtl, "constant", emission_part.inputs[0], -1)					
				
		# Export normal if found :
		if normal_map_input:

			if normal_map_input.is_linked:
				nmap_input = normal_map_input.links[0].from_node
				
				# Check normal_map_input is a "Normal Map" node.
				# Blender supports a different normal map on specular & diffuse. XSG doesn't. We just check diffuse normal map & use that for both.
				if nmap_input.type == 'NORMAL_MAP' :
				
					if nmap_input.inputs[1].is_linked :
						# Range is 0 .. 10 in blender . 1 seems to mean full normal map without saturation. 
						# xsg also supports -ve value to invert normal map.
						level = nmap_input.inputs[0].default_value 
						Export_Material.Part_Write(exp, mtl, "normal", nmap_input.inputs[1], level)
				
				
				
		# TODO: Merge specular channel export into Part_Write. 
		# Need to add a flag to pick up and export gloss in addition to standard parameters.
		# Blender's range is 1 - 511. xsg is 0-1
		#Specularity = 1000 * (mtl.specular_hardness - 1.0) / 510.0
		#Specular = list(Vector(mtl.specular_color) * mtl.specular_intensity)
		#file.Write('<part id="specular" gloss={:f}><input color="{:f} {:f} {:f}"/></>\n'.format(Specularity, Specular[0], Specular[1], Specular[2]))
		
		exp.file.Unindent()
		
		if not mtl_open_complete:
			exp.file.Write('>', Indent=False)
			mtl_open_complete = True
		
		exp.file.Write("</material>\n");
		exp.file.Unindent()
		
	@staticmethod
	def GetVectorInput(node):
		for inp in node.inputs:
			if inp.is_linked and inp.type == 'VECTOR' :
				l = inp.links[0]
				link_src = l.from_node
				return link_src
		return None
		
	@staticmethod
	def Map_Write(exp, node, color, element_open_required):
		
		Export_Material.Input_Begin(exp, node, element_open_required)   

		if color:
			exp.file.Write(' color="{:f} {:f} {:f}"'.format(color[0], color[1], color[2]), Indent=False)
		
        # TODO: if first char is '.', prepend current directory.
		texture_path = bpy.path.abspath(node.image.filepath); #, exp.src_dir)
		
		texture_filename = bpy.path.basename(texture_path)
		
		# Copy source texture file to output directory if we can find it
		if os.path.isfile(texture_path) :
			dest_path = os.path.join(exp.texture_path, texture_filename)
			shutil.copyfile(texture_path, dest_path)
		else:
			texture_filename = None   # Remove unresolved textures for now.
		#	default_source = bpy.utils.user_resource('SCRIPTS', "addons")
		#	default_source += "/io_scene_xsg/undefined.png"
		#
		#	if os.path.isfile(default_source) :
		#		shutil.copyfile(default_source, dest_path)


		# infer xsg input type from file extension
		if texture_filename:
			notused_file_name, file_extension = os.path.splitext(texture_filename)
			if file_extension == '.mp4':
				exp.file.Write(' type="video"', Indent=False)
		
		vinput = Export_Material.GetVectorInput(node)
		uv_map = None
		
		scale = Vector((1,1,1))
		
		if vinput:
			if vinput.type == 'MAPPING':
				
				if vinput.vector_type == 'POINT' or vinput.vector_type == 'VECTOR' :
					scale_node = vinput.inputs['Scale']
					if scale_node.is_linked:
						l = scale_node.links[0]
						link_src = l.from_node
						svalue = link_src.outputs[0].default_value
						scale = Vector((svalue, svalue, svalue))
					else:
						scale = vinput.inputs['Scale'].default_value
				else:
					print("ignoring unsupported mapping vector_type={}".format(vinput.vector_type))
				
				# Grab mapping input ...
				vinput = Export_Material.GetVectorInput(vinput)
			elif vinput.type != 'UVMAP':  # handled below
				exp.Log("<><> Unsupported vinput_1.type={}".format(vinput.type), Error)
				
		if vinput:
			if vinput.type == 'UVMAP':
				if vinput.uv_map != "UVMap":
					uv_map = vinput.uv_map
			elif vinput.type == 'TEX_COORD':
				# Select default uv_map
				uv_map = None
			else:
				exp.Log("<><> Warning: Unsupported MAPPING Vector input type={}".format(vinput.type), Warning)
		
		if texture_filename:
			exp.file.Write(' scale="{} {}" src="{}"'.format(scale[0], -scale[1], texture_filename), Indent=False)
		
		if node.extension == 'CLIP':
			exp.file.Write(' addr="rclamp"', Indent=False)  # reverse clamp in XSG (0 thru -1) because must currently flip images upside down to match mapping.
		
		# Gets uv map channel name. XSG currently only supports texture coordinate set id as index reference into mesh tcoord sets.
		# Up to two sets are currently supported by Infinity. Hence anything other than the default is set 1. (default is set 0)

		#print("UV={}".format(uv_map))

		if uv_map and uv_map != "0":
			exp.file.Write(' tcoord="{}"'.format(uv_map), Indent=False)  # Actual ID is : "{}"'.format(uv_map)
			
		if element_open_required:
			exp.file.Write("/>\n", Indent=False)
			
	
	def Input_Begin(exp, node, element_open_required = True):

		if element_open_required:
			exp.file.Write('<input')

		if hasattr(node, 'label'):
			index = node.label.find('xsg.input:')
			if index == 0:
				exp.file.Write(node.label[10:], Indent=False) # Write the custom node property
				exp.file.Write(' ', Indent=False)
				

	def RGB_Write(exp, node):
		Export_Material.Input_Begin(exp, node)
		color = node.outputs[0].default_value
		exp.file.Write(' color="{:f} {:f} {:f}"/>\n'.format(color[0], color[1], color[2]), Indent=False)
		

	def HueSat_Write(exp, node):

		node_input = node.inputs['Color']

		if node_input.is_linked:
			Export_Material.Input_Write(exp, node_input.links[0].from_node, None)
		else:
			Export_Material.Input_Begin(exp, node)
			color = node_input.default_value
			exp.file.Write(' color="{:f} {:f} {:f}"/>\n'.format(color[0], color[1], color[2]), Indent=False)
			

	def Group_Write(exp, node):

		# Currently hardwired Group - assumes 3 inputs to the mixmap (R,G,B) followed by 3 inputs to the mixmap inputs.
		# For anything else we'll have to identify somehow & evolve.

		exp.Log("Exporting group : {} / {}".format(node.label, node.node_tree.name), Info)
		
		node_inputs = [ None, None, None, None ]
		node_mixer = None
		
		for inp in node.inputs:
			exp.Log(inp.name)
			
			if inp.name == "R":
				node_mixer = inp
				
			if inp.name == "Input_R":
				node_inputs[0] = inp

			if inp.name == "Input_G":
				node_inputs[1] = inp

			if inp.name == "Input_B":
				node_inputs[2] = inp
				
			if inp.name == "Input_A":
				node_inputs[3] = inp

		
		# Wire up mixer inputs.			
		#exp.dump(node_mixer)

		if node_mixer and node_mixer.is_linked:
			exp.file.Write('<mix')
			Export_Material.Input_Write(exp, node_mixer.links[0].from_node, None, False)
			exp.file.Write('>\n', False)
			
			exp.file.Indent()

			for i in range(3, 6):
				node_input = node.inputs[i]
				if node_input.is_linked:		
					Export_Material.Input_Write(exp, node_input.links[0].from_node, None)
			
			exp.file.Unindent()

			exp.file.Write('</mix>\n')
		else:
			exp.Log("Unknown mixer config [ignored]", Warning)
			
		
	def Input_Write(exp, src, color, element_open_required = True) :
	
		if src.type == 'TEX_IMAGE':
			Export_Material.Map_Write(exp, src, color, element_open_required)
		elif src.type == 'MIX':
			Export_Material.Mixer_Write(exp, src, src.blend_type)
		elif src.type == 'RGB':
			Export_Material.RGB_Write(exp, src)
		elif src.type == 'HUE_SAT':
			Export_Material.HueSat_Write(exp, src)
		elif src.type == 'GROUP':
			Export_Material.Group_Write(exp, src)
		elif src.type == 'SEPARATE_COLOR':
			if src.inputs[0].is_linked:	# Just pass straight thru for now.
				Export_Material.Input_Write(exp, src.inputs[0].links[0].from_node, color, element_open_required)
		elif src.type == 'INVERT':
			Export_Material.Invert_Write(exp, src, color)
		elif src.type == 'RGBTOBW' :
			if src.inputs[0].is_linked:
				Export_Material.Input_Write(exp, src.inputs[0].links[0].from_node, False, element_open_required)
		else:
			exp.Log("Unsupported input node : {}".format(src.type), Warning)
			Export_Material.Input_Begin(exp, src)
			exp.file.Write('type="' + src.type + '"/>\n')
			
			
	# Implemented same way as Mixer_Write, only with type=inv. TODO: Factor
	def Invert_Write(exp, src, color) :	
		
		exp.file.Write('<mix type="inv">\n')
		exp.file.Indent()
		
		for i in src.inputs:
			if i.is_linked:
			
				input = i.links[0].from_node

				if i.name != 'Factor':  # ignore factor input for now.
					Export_Material.Input_Write(exp, input, None)
					
		exp.file.Unindent()				
		exp.file.Write('</mix>\n')
		
		
	def Mixer_Write(exp, src, blend_type) :
	
		# Inputs: Factor is in 0, RGB inputs are 1 & 2.
	
		# Write simplistic modulate mixer in compact form
		if blend_type == 'MULTIPLY':
			if src.inputs[1].links[0].from_node.type == "RGB":
				Export_Material.Input_Write(exp, src.inputs[2].links[0].from_node, src.inputs[1].links[0].from_node.outputs[0].default_value)
			elif src.inputs[2].links[0].from_node.type == "RGB":
				Export_Material.Input_Write(exp, src.inputs[1].links[0].from_node, src.inputs[2].links[0].from_node.outputs[0].default_value)
		else:
			exp.file.Write('<mix>\n')
			exp.file.Indent()
		
			for i in src.inputs:
				if i.is_linked:
				
					input = i.links[0].from_node

					if i.name != 'Factor':  # ignore factor input for now.
						Export_Material.Input_Write(exp, input, None)
					
			exp.file.Unindent()				
			exp.file.Write('</mix>\n')
			
		
	@staticmethod
	def Part_Write(exp, mtl, inf_id, node_input, level) :

		exp.file.Write('<part id="')
		exp.file.Write(inf_id, Indent=False)
		exp.file.Write('"', Indent=False)
		
		if level >= 0.0 :
			exp.file.Write(' level="{}"'.format(level), Indent=False)
		
		exp.file.Write('>\n', Indent=False)			
		
		exp.file.Indent()
		
		if node_input.is_linked:
			Export_Material.Input_Write(exp, node_input.links[0].from_node, None)
		else:
			Export_Material.Input_Begin(exp, node_input.node)
			exp.file.Write(' color="{:f} {:f} {:f}"'.format(node_input.default_value[0], node_input.default_value[1], node_input.default_value[2]), Indent=False)
			exp.file.Write(' ', Indent=False)
			exp.file.Write('/>\n', Indent=False)
	
		exp.file.Unindent()
		exp.file.Write('</part>\n')		